
#include "stdafx.h"

#include "editor_base.h"


using namespace bec;


RefreshCentry::RefreshCentry(BaseEditor &editor) : _editor(editor)
{
  _editor.block_auto_refresh();
};

RefreshCentry::~RefreshCentry()
{
  _editor.unblock_auto_refresh();
}


BaseEditor::BaseEditor(GRTManager *grtm, const grt::Ref<GrtObject> &object)
: _grtm(grtm), _object(object), _locked(false)
{
  if (object.is_valid())
    add_listeners(object);
}


BaseEditor::~BaseEditor()
{
}


void BaseEditor::add_listeners(const grt::Ref<GrtObject> &object)
{
  object->signal_changed().connect(sigc::mem_fun(this, &BaseEditor::object_member_changed));
  object->signal_list_changed().connect(sigc::hide(sigc::hide(sigc::hide(sigc::mem_fun(this, &BaseEditor::on_object_changed)))));
}


void BaseEditor::object_member_changed(const std::string &member, const grt::ValueRef &ovalue)
{
  if (_ignored_object_fields_for_ui_refresh.find(member) == _ignored_object_fields_for_ui_refresh.end())    
    on_object_changed();
}


void BaseEditor::on_object_changed()
{
  if (_refresh_ui)
    _refresh_ui();
}

bool BaseEditor::begin_undo_object_text_change(const grt::ObjectRef &object, const std::string &member)
{
  grt::UndoAction *action= get_undo_manager()->get_latest_undo_action();
  grt::UndoGroup *group;
  grt::UndoObjectChangeAction* taction;
  
  if ((group= dynamic_cast<grt::UndoGroup*>(action)))
  {
    if (group->get_actions().empty())
      taction= 0;
    else
      taction= dynamic_cast<grt::UndoObjectChangeAction*>(group->get_actions().front());
  }
  else
    taction= dynamic_cast<grt::UndoObjectChangeAction*>(action);

  // if the previous action was of the same type for the same object, then don't add a new undo
  // action, since the previous one will do the same thing
  if (taction)
  {
    if (taction->get_object() == object && taction->get_member() == member)
      return false;
  }

  begin_undo_group();
  action= new grt::UndoObjectChangeAction(object, member);
  action->signal_apply().connect(sigc::mem_fun(this, &BaseEditor::undo_applied));

  get_undo_manager()->add_undo(action);

  return true;
}


bool BaseEditor::add_undo_object_text_change(const grt::ObjectRef &object, const std::string &member)
{
  grt::UndoAction *action= get_undo_manager()->get_latest_undo_action();
  grt::UndoGroup *group;
  grt::UndoObjectChangeAction* taction;
  
  if ((group= dynamic_cast<grt::UndoGroup*>(action)))
    taction= dynamic_cast<grt::UndoObjectChangeAction*>(group->get_actions().front());
  else
    taction= dynamic_cast<grt::UndoObjectChangeAction*>(action);

  // if the previous action was of the same type for the same object, then don't add a new undo
  // action, since the previous one will do the same thing
  if (taction)
  {
    if (taction->get_object() == object && taction->get_member() == member)
      return false;
  }

  action= new grt::UndoObjectChangeAction(object, member);
  action->signal_apply().connect(sigc::mem_fun(this, &BaseEditor::undo_applied));

  get_undo_manager()->add_undo(action);

  return true;
}


void BaseEditor::add_undo(grt::UndoAction *action, const std::string &description)
{
  action->signal_apply().connect(sigc::mem_fun(this, &BaseEditor::undo_applied));
  get_undo_manager()->add_undo(action);
  if (!description.empty())
    get_undo_manager()->set_action_description(description);
}


void BaseEditor::begin_undo_group()
{
  get_undo_manager()->begin_undo_group();
}


void BaseEditor::end_undo_group(const std::string &description)
{
  get_undo_manager()->end_undo_group();
  get_undo_manager()->set_action_description(description);
}


void BaseEditor::cancel_undo_group()
{
  get_undo_manager()->cancel_undo_group();
}


void BaseEditor::set_refresh_ui_slot(const sigc::slot<void> &slot)
{
  _refresh_ui= slot;
}



void BaseEditor::undo_applied()
{
  _grtm->run_when_idle(sigc::bind_return<bool>(_refresh_ui, false));
}



void BaseEditor::run_from_grt(const sigc::slot<void> &slot)
{
  GRTDispatcher *disp= _grtm->get_dispatcher();

  disp->execute_simple_function("editor action",
    sigc::bind_return<grt::ValueRef>(sigc::hide(slot), grt::ValueRef()));
}


/** 
 ****************************************************************************
 * @brief Marks the editor as in use by the frontend.
 *
 * When the frontend starts displaying data from the editor backend,
 * it should lock it and when it's done with it, it should unlock it.
 * If the editor backend is unlocked and there are no unapplied changes to it,
 * it will emit a "unlocked" signal, which will normally release the backend.
 *
 *****************************************************************************/
void BaseEditor::lock()
{
  _locked= true;
}


void BaseEditor::unlock()
{
  _locked= false;

  _unlocked_signal.emit(this);
}

