#include "stdafx.h"

#include "model_figure_impl.h"

#include "workbench_physical_view_impl.h"
#include "workbench_physical_tablefigure_impl.h"

#include <grtpp_undo_manager.h>

using namespace wbbridge::physical;


workbench_physical_View::ImplData::ImplData(workbench_physical_View *owner)
  : super(owner)
{
}


workbench_physical_LayerRef workbench_physical_View::ImplData::place_new_layer(double x, double y, double width, double height, const std::string &name)
{
  grt::UndoManager *um= 0;
  workbench_physical_LayerRef layer(self()->get_grt());

  if (self()->is_global() && !self()->get_grt()->tracking_changes())
    um= self()->get_grt()->get_undo_manager();

  if (um)
    um->begin_undo_group();

  layer->owner(self());
  layer->left(x);
  layer->top(y);
  layer->width(width);
  layer->height(height);
  layer->name(name);

  self()->_layers.insert(layer);

  model_LayerRef rootLayer(self()->rootLayer());
  
  // capture items inside the layer area
  mdc::Rect bounds(x, y, width, height);
  // check for objects in the parent layer that fall inside this layer and add them
  for (size_t c= rootLayer->figures().count(), i= 0; i < c; i++)
  {
    size_t idx= c-i-1;
    model_FigureRef fig(rootLayer->figures().get(idx));
    mdc::Rect fbounds(*fig->left(), *fig->top(), *fig->width(), *fig->height());

    if (mdc::bounds_contain_bounds(bounds, fbounds))
    {
      if (um)
        um->add_undo(new grt::UndoObjectChangeAction(fig, "layer"));
      fig->layer(layer);

      if (um)
        um->add_undo(new grt::UndoListRemoveAction(rootLayer->figures(), idx));
      rootLayer->figures().remove(idx);

      if (um)
        um->add_undo(new grt::UndoListInsertAction(layer->figures()));
      layer->figures().insert(fig);
    }
  }

  if (um)
    um->end_undo_group(strfmt(_("Place '%s'"), name.c_str()));

  return layer;
}


workbench_physical_TableFigureRef workbench_physical_View::ImplData::place_table(const db_TableRef &table, double x, double y)
{
  workbench_physical_TableFigureRef figure(self()->get_grt());
  
  figure->owner(self());
  figure->table(table);
  figure->left(x);
  figure->top(y);
  figure->layer(self()->rootLayer());
  figure->name(table->name());

  self()->addFigure(figure);

  create_connections_for_table(table);

  return figure;
}


workbench_physical_RoutineGroupFigureRef workbench_physical_View::ImplData::place_routine_group(const db_RoutineGroupRef &rgroup, double x, double y)
{
  workbench_physical_RoutineGroupFigureRef figure(self()->get_grt());

  figure->owner(self());
  figure->routineGroup(rgroup);
  figure->left(x);
  figure->top(y);
  figure->layer(self()->rootLayer());
  figure->name(rgroup->name());

  self()->addFigure(figure);

  return figure;
}


workbench_physical_ViewFigureRef workbench_physical_View::ImplData::place_view(const db_ViewRef &view, double x, double y)
{
  workbench_physical_ViewFigureRef figure(self()->get_grt());

  figure->owner(self());
  figure->view(view);
  figure->left(x);
  figure->top(y);
  figure->layer(self()->rootLayer());
  figure->name(view->name());

  self()->addFigure(figure);

  return figure;
}


void workbench_physical_View::ImplData::notify_table_realized(workbench_physical_TableFigure *table)
{
  wbfig::Table *t= dynamic_cast<wbfig::Table*>(table->get_data()->get_canvas_item());

  t->signal_enter().connect(sigc::bind<workbench_physical_TableFigure*>(sigc::mem_fun(this, &workbench_physical_View::ImplData::enter_table_callback),table));
  t->signal_leave().connect(sigc::bind<workbench_physical_TableFigure*>(sigc::mem_fun(this, &workbench_physical_View::ImplData::leave_table_callback),table));
}


void workbench_physical_View::ImplData::enter_table_callback(workbench_physical_TableFigure* table)
{
  _enter_slot(table);
}


void workbench_physical_View::ImplData::leave_table_callback(workbench_physical_TableFigure* table)
{
  _leave_slot(table);
}


void workbench_physical_View::ImplData::set_table_enter_slot(const sigc::slot<void, workbench_physical_TableFigure*> &slot)
{
  _enter_slot= slot;
}


void workbench_physical_View::ImplData::set_table_leave_slot(const sigc::slot<void, workbench_physical_TableFigure*> &slot)
{
  _leave_slot= slot;
}




model_FigureRef workbench_physical_View::ImplData::get_figure_for_dbobject(const db_DatabaseObjectRef &obj)
{
  if (obj.is_valid())
  {
    std::map<std::string, model_FigureRef>::iterator iter;

    iter= _dbobject_to_figure.find(obj.id());

    if (iter != _dbobject_to_figure.end())
      return iter->second;
  }
  return model_FigureRef();
}



void workbench_physical_View::ImplData::add_mapping(const db_DatabaseObjectRef &object, const model_FigureRef &figure)
{
  _dbobject_to_figure[object.id()]= figure;
}


void workbench_physical_View::ImplData::remove_mapping(const db_DatabaseObjectRef &object)
{
  _dbobject_to_figure.erase(object.id());
}



workbench_physical_ConnectionRef workbench_physical_View::ImplData::create_connection_for_foreign_key(const db_ForeignKeyRef &fk)
{
  // check if both tables referenced in fk are in the view
  if (_fk_to_connection.find(fk.id()) == _fk_to_connection.end()
      && get_figure_for_dbobject(db_DatabaseObjectRef::cast_from(fk->owner())).is_valid()
      && get_figure_for_dbobject(fk->referencedTable()).is_valid())
  {
    workbench_physical_ConnectionRef conn(self()->get_grt());

    conn->owner(self());
    conn->name("");
    conn->caption(fk->name());
    // start and end figures are set by the internal code from the connection
    conn->foreignKey(fk);

    self()->addConnection(conn);

    return conn;
  }
  return workbench_physical_ConnectionRef();
}


/** Creates or updates connections for a table.

This will create connection objects for each foreign key in the table if
both referencing and referenced tables are in the view. It will also check if 
the table is referenced by an existing table's foreign key and create connections
if possible.
 */
int workbench_physical_View::ImplData::create_connections_for_table(const db_TableRef &table)
{    
  int c= 0;

  if (table.is_valid())
  {
    // first create connections for FKs from the table
    for (grt::ListRef<db_ForeignKey>::const_iterator 
         end= table->foreignKeys().end(), fk= table->foreignKeys().begin(); fk != end; ++fk)
    {
      if (create_connection_for_foreign_key(*fk).is_valid())
        c++;
    }

    // create connections for FKs that reference this one
    db_SchemaRef schema(db_SchemaRef::cast_from(table->owner()));
    
    if (schema.is_valid())
    {
      grt::ListRef<db_ForeignKey> fks(schema->getForeignKeysReferencingTable(table));

      for (grt::ListRef<db_ForeignKey>::const_iterator fk= fks.begin(); fk != fks.end(); ++fk)
      {
        if (create_connection_for_foreign_key(*fk).is_valid())
          c++;
      }
    }
  }

  return c;
}


void workbench_physical_View::ImplData::delete_connections_for_table(const db_TableRef &table)
{
  if (table.is_valid())
  {
    // first delete connections for FKs from the table
    for (grt::ListRef<db_ForeignKey>::const_iterator 
         end= table->foreignKeys().end(), fk= table->foreignKeys().begin(); fk != end; ++fk)
    {
      workbench_physical_ConnectionRef conn(get_connection_for_foreign_key(*fk));
      if (conn.is_valid())
        remove_connection(conn);
    }

    // delete connections for FKs that reference this one
    db_SchemaRef schema(db_SchemaRef::cast_from(table->owner()));

    if (schema.is_valid())
    {
      grt::ListRef<db_ForeignKey> fks(schema->getForeignKeysReferencingTable(table));
      
      for (grt::ListRef<db_ForeignKey>::const_iterator fk= fks.begin(); fk != fks.end(); ++fk)
      {
        workbench_physical_ConnectionRef conn(get_connection_for_foreign_key(*fk));
        if (conn.is_valid())
          remove_connection(conn);
      }
    }
  }
}


workbench_physical_ConnectionRef workbench_physical_View::ImplData::get_connection_for_foreign_key(const db_ForeignKeyRef &fk)
{
  std::map<std::string, workbench_physical_ConnectionRef>::iterator iter;

  iter= _fk_to_connection.find(fk.id());

  if (iter != _fk_to_connection.end())
    return iter->second;

  return workbench_physical_ConnectionRef();
}


void workbench_physical_View::ImplData::add_fk_mapping(const db_ForeignKeyRef &fk, const workbench_physical_ConnectionRef &connection)
{
  _fk_to_connection[fk.id()]= connection;
}


void workbench_physical_View::ImplData::remove_fk_mapping(const db_ForeignKeyRef &fk)
{
  _fk_to_connection.erase(fk.id());
}


void workbench_physical_View::ImplData::add_connection(const model_ConnectionRef &conn)
{
  model_View::ImplData::add_connection(conn);
  add_fk_mapping(workbench_physical_ConnectionRef::cast_from(conn)->foreignKey(), workbench_physical_ConnectionRef::cast_from(conn));
}


void workbench_physical_View::ImplData::remove_connection(const model_ConnectionRef &conn)
{
  model_View::ImplData::remove_connection(conn);
  remove_fk_mapping(workbench_physical_ConnectionRef::cast_from(conn)->foreignKey());
}




void workbench_physical_View::ImplData::auto_place_db_objects(const grt::ListRef<db_DatabaseObject> &objects)
{
  grt::GRT *grt= self()->get_grt();
  grt::Module *module= grt->get_module("WbModel");

  grt::BaseListRef args(grt);
  grt::DictRef options(grt);

  options.set("selected_objects", objects);

  args.ginsert(workbench_physical_ModelRef::cast_from(self()->owner())->catalog());
  args.ginsert(options);

  module->call_function("autoplace_ex", args);
}
