#include "stdafx.h"

#include "wb_model_view_form.h"
#include "wb_context.h"
#include "wb_component.h"
#include "canvas_floater.h"

#include "mdc_back_layer.h"

#include "grt/common.h"
#include "grt/clipboard.h"

#include "wbcanvas/model_figure_impl.h"
#include "wbcanvas/model_layer_impl.h"
#include "wbcanvas/model_view_impl.h"
#include "wbcanvas/model_connection_impl.h"

#include "wb_context_ui.h"

#define DEFAULT_TOOL "basic/select"

#define TOOLTIP_DELAY 1.0

using namespace wb;
using namespace bec;


class wb::Tooltip : public mdc::Box
{
  mdc::TextFigure text;

public:
  Tooltip(mdc::Layer *layer)
    : mdc::Box(layer), text(layer)
  {
    set_padding(4, 4);

    set_draw_background(true);
    set_background_color(mdc::Color(1, 1, 0.8));
    set_border_color(mdc::Color(0.5, 0.5, 0.5));

    mdc::FontSpec font(text.get_font());
#ifdef _WIN32
    font.family= "Verdana";
#else
    font.family= "Bitstream Vera Sans";
#endif
    font.size= 11;
    text.set_font(font);
    text.set_pen_color(mdc::Color(0.3, 0.3, 0.3));
    text.set_multi_line(true);

    add(&text, true, true);
  }

  void set_text(const std::string &text)
  {
    this->text.set_text(text);
  }
};


ModelViewForm::ModelViewForm(WBComponent *owner, const model_ViewRef &view)
: _view(0), _owner(owner), _model_view(view)
{
  _drag_panning= false;
  _space_panning= false;

  _last_over_item= 0;
  _tooltip_timer= 0;
  _tooltip= 0;

  _main_layer= 0;
  _floater_layer= 0;

  _tool= DEFAULT_TOOL;

  _paste_offset= 0;

  _latest_drag_action= 0;

  _shortcuts= owner->get_wb()->get_ui()->get_command_ui()->get_shortcuts_for_context(WB_CONTEXT_MODEL);

  owner->get_wb()->get_clipboard()->signal_changed().connect(sigc::mem_fun(this, &ModelViewForm::clipboard_changed));
}


ModelViewForm::~ModelViewForm()
{
}


void ModelViewForm::attach_canvas_view(mdc::CanvasView *cview)
{
  _view= cview;

  cview->set_tag(_model_view.id());
  cview->set_grid_snapping(_model_view->options().get_int("AlignToGrid", 0)!=0);
  cview->get_background_layer()->set_grid_visible(_model_view->options().get_int("ShowGrid", 1)!=0);

  cview->get_selection()->signal_begin_dragging().connect(sigc::mem_fun(this, &ModelViewForm::begin_selection_drag));
  cview->get_selection()->signal_end_dragging().connect(sigc::mem_fun(this, &ModelViewForm::end_selection_drag));

  _main_layer= _view->get_current_layer();
  _floater_layer= _view->new_layer("floater");
}


void ModelViewForm::set_button_callback(const sigc::slot<bool, ModelViewForm*, mdc::MouseButton, bool, mdc::Point, mdc::EventState> &cb)
{
  _handle_button= cb;
}


void ModelViewForm::set_motion_callback(const sigc::slot<bool, ModelViewForm*, mdc::Point, mdc::EventState> &cb)
{
  _handle_motion= cb;
}


void ModelViewForm::set_reset_tool_callback(const sigc::slot<void, ModelViewForm*> &cb)
{
  _reset_tool= cb;
}


void ModelViewForm::enable_panning(bool flag)
{
  if (flag)
  {
    _old_tool= _tool;
    _old_reset_tool= _reset_tool;
    _old_handle_button= _handle_button;
    _old_handle_motion= _handle_motion;
    _old_cursor= _cursor;
    _tool= "basic/hand";
    WBComponent *compo= _owner->get_wb()->get_component_named("basic");
    compo->setup_canvas_tool(this, _tool);

    _owner->get_wb()->tool_changed(_view);
  }
  else
  {
    _reset_tool(this);
    _tool= _old_tool;
    _cursor= _old_cursor;
    _reset_tool= _old_reset_tool;
    _handle_button= _old_handle_button;
    _handle_motion= _old_handle_motion;
    _owner->get_wb()->tool_changed(_view);
  }
}


void ModelViewForm::begin_selection_drag()
{
  // check if old selection is the same as the new

  mdc::Selection::ContentType selection(_view->get_selection()->get_contents());
  bool mismatch= false;

  if (!_latest_drag_action || 
    _latest_drag_action != _owner->get_wb()->get_undo_manager()->get_latest_undo_action())
    mismatch= true;
  else
  {
    if (selection.size() != _old_positions.size())
      mismatch= true;
    else
    {
      for (std::map<grt::internal::Value*,mdc::Point>::iterator iter= _old_positions.begin();
        iter != _old_positions.end(); ++iter)
      {
        grt::ValueRef value(iter->first);
        
        if (model_FigureRef::can_wrap(value)
            && selection.find(model_FigureRef::cast_from(value)->get_data()->get_canvas_item()) == selection.end())
        {
          mismatch= true;
          break;
        }
        else if (model_LayerRef::can_wrap(value) 
                 && selection.find(model_LayerRef::cast_from(value)->get_data()->get_area_group()) == selection.end())
        {
          mismatch= true;
          break;
        }
      }
    }
  }

  mdc::AreaGroup *root= _view->get_current_layer()->get_root_area_group();
  // remove objects from respective layers so they can be freely dragged
  for (mdc::Selection::ContentType::const_iterator iter= selection.begin();
    iter != selection.end(); ++iter)
  {
    mdc::Point rpos= (*iter)->get_root_position();

    if (root != (*iter)->get_parent() && !(*iter)->get_parent()->get_selected())
    {
      root->add(*iter);
      (*iter)->move_to(rpos);
    }
  }

  if (mismatch)
  {
    _old_positions.clear();

    grt::ListRef<model_Object> sel(_model_view->selection());
    for (size_t c= sel.count(), i= 0; i < c; i++)
    {
      if (sel[i]->is_instance(model_Figure::static_class_name()))
      {
        model_FigureRef figure(model_FigureRef::cast_from(sel[i]));

        _old_positions[figure.valueptr()]= mdc::Point(figure->left(), figure->top());
      }
      else if (sel[i]->is_instance(model_Layer::static_class_name()))
      {
        model_LayerRef layer(model_LayerRef::cast_from(sel[i]));

        _old_positions[layer.valueptr()]= mdc::Point(layer->left(), layer->top());
      }
    }

    // signal that we need to record undo info when drag is over
    _latest_drag_action= 0;
  }
}


void ModelViewForm::end_selection_drag()
{
  grt::UndoManager *um= _owner->get_wb()->get_undo_manager();

  std::string name;
  int count= 0;

  // _latest_drag_action is not NULL we can compress this move event from
  // the undo stack, since the same objects were moved. except if the movement
  // caused it to be dragged into/out of a layer
  um->begin_undo_group();

  grt::ListRef<model_Object> sel(_model_view->selection());
  for (size_t c= sel.count(), i= 0; i < c; i++)
  {
    if (sel[i]->is_instance(model_Figure::static_class_name()))
    {
      model_FigureRef figure(model_FigureRef::cast_from(sel[i]));
      if (!_latest_drag_action)
      {
        mdc::Point p= _old_positions[figure.valueptr()];

        um->add_undo(new grt::UndoObjectChangeAction(figure, "left", grt::DoubleRef(p.x)));
        um->add_undo(new grt::UndoObjectChangeAction(figure, "top", grt::DoubleRef(p.y)));
      }
      name= figure->name();
      count++;
    }
    else if (sel[i]->is_instance(model_Layer::static_class_name()))
    {
      model_LayerRef layer(model_LayerRef::cast_from(sel[i]));
      if (!_latest_drag_action)
      {
        mdc::Point p= _old_positions[layer.valueptr()];

        um->add_undo(new grt::UndoObjectChangeAction(layer, "left", grt::DoubleRef(p.x)));
        um->add_undo(new grt::UndoObjectChangeAction(layer, "top", grt::DoubleRef(p.y)));
      }
      name= layer->name();
      count++;
    }
  }

  if (!_latest_drag_action)
    relocate_selection();
  else
  {
    if (!relocate_selection())
    {
      um->cancel_undo_group();
      return;
    }
  }

  if (!name.empty() && count == 1)
    um->end_undo_group(fmt(_("Move %s"), name.c_str()));
  else
    um->end_undo_group(_("Move Objects"));

  _latest_drag_action= um->get_latest_undo_action();
}


void ModelViewForm::handle_mouse_button(mdc::MouseButton button, bool press, int x, int y, mdc::EventState state)
{
  mdc::Point pos(_view->window_to_canvas(x, y));

  if (button == mdc::ButtonRight && !press)
  {
    model_ObjectRef object= get_object_at(pos);

    bec::MenuItem item;
    bec::MenuItemList items;

    // if the user clicks in a selected object or in the view just show popup 
    // if user clicks in an unseleced object, select it and show popup

    if (object.is_valid() && _model_view->selection().get_index(object) == grt::BaseListRef::npos)
      _view->get_selection()->set(_view->get_item_at(pos));

    {
      std::list<std::string> groups;

      groups.push_back("Catalog/*");
      groups.push_back("Model/*");

      _owner->get_wb()->get_object_list_popup_items(this, grt::ListRef<GrtObject>::cast_from(_model_view->selection()),
        get_edit_target_name(), groups, items);
    }
    //else
      //_owner->get_model_popup_items(items);

    if (!items.empty())
    {
      int x, y;
      _view->canvas_to_window(pos, x, y);
      _owner->get_wb()->popup_menu(&items, x, y);
    }
    return;
  }

  // temporarily select Hand tool when middle-button-dragging
  if (button == mdc::ButtonMiddle)
  {
    if (press && !_drag_panning && !_space_panning)
    {
      _drag_panning= true;
      enable_panning(true);
    }
    else if (!press && _drag_panning)
    {
      _drag_panning= false;
      enable_panning(false);
    }
  }

  if (button != mdc::ButtonLeft && button != mdc::ButtonMiddle)
    return;

  if (_handle_button && _handle_button(this, button, press, pos, state))
    return;

  
  _view->handle_mouse_button(button, press, x, y, state);
}


void ModelViewForm::handle_mouse_move(int x, int y, mdc::EventState state)
{
  mdc::Point pos(_view->window_to_canvas(x, y));

  if (_handle_motion && _handle_motion(this, pos, state))
    return;

  _view->handle_mouse_move(x, y, state);

  if (state == 0)
  {
    mdc::CanvasItem *over= _view->get_leaf_item_at(pos);

    if (over != _last_over_item)
    {
      if (_tooltip_timer)
      {
        _owner->get_grt_manager()->cancel_timer(_tooltip_timer);
        _tooltip_timer= 0;
      }

      if (_tooltip && _tooltip->get_visible())
      {
        _tooltip->set_visible(false);
        _view->set_needs_repaint();
      }
      if (_tooltip && _tooltip->get_visible() && over)
        show_tooltip(pos);
      else
      {
        // check if there's something under pointer
        if (get_object_at(pos).is_valid())
        {
          _tooltip_timer= _owner->get_grt_manager()->run_every(
            sigc::bind<mdc::Point>(sigc::mem_fun(this, &ModelViewForm::show_tooltip), pos),
            TOOLTIP_DELAY);
        }
      }

      //if (_last_over_item)
      //  hover_out_figure(_last_over_item);

      _last_over_item= over;

      //hover_in_figure(over, pos);
    }
  }
  else
  {
    if (_tooltip_timer)
    {
      _owner->get_grt_manager()->cancel_timer(_tooltip_timer);
      _tooltip_timer= 0;
    }

    if (_tooltip && _tooltip->get_visible())
    {
      _tooltip->set_visible(false);
      _view->set_needs_repaint();
    }
  }
}


void ModelViewForm::handle_key(const mdc::KeyInfo &key, bool press, mdc::EventState state)
{
  if (press)
  {    
    for (std::vector<WBShortcut>::const_iterator iter= _shortcuts.begin();
      iter != _shortcuts.end(); ++iter)
    {
      if (iter->modifiers == state && iter->key == key)
      {
        _owner->get_wb()->get_ui()->get_command_ui()->activate_command(iter->command);
        return;
      }
    }
  }
  if (key.keycode == mdc::KSpace)
  {
    if (press && !_drag_panning && !_space_panning)
    {
      _space_panning= true;
      enable_panning(true);
    }
    else if (!press && _space_panning)
    {
      _space_panning= false;
      enable_panning(false);
    }
  }

  _view->handle_key(key, press, state);
}



bool ModelViewForm::show_tooltip(const mdc::Point &pos)
{
  model_ObjectRef object(get_object_at(pos));

  if (object.is_valid())
  {
    std::string text= _owner->get_wb()->get_object_tooltip(object, pos);
    if (!text.empty())
    {
      if (!_tooltip)
      {
        _tooltip= new Tooltip(_view->get_interaction_layer());

        _view->get_interaction_layer()->add_item(_tooltip);
      }

      _tooltip->set_text(text);
      _tooltip->move_to(pos + mdc::Point(0, 20));
      _tooltip->set_visible(true);

      _view->get_interaction_layer()->get_root_area_group()->raise_item(_tooltip);
    }
  }

  return false;
}




void ModelViewForm::set_tool(const std::string &tool)
{
  reset_tool(false);

  _tool= tool;

  WBComponent *compo= _owner->get_wb()->get_component_named(split_string(tool, "/")[0]);
  if (compo)
    compo->setup_canvas_tool(this, tool);
  else
    throw std::runtime_error("Invalid tool "+tool);

  _owner->get_wb()->tool_changed(_view);
}


void ModelViewForm::reset_tool(bool notify)
{
 // _tool_args.clear();

  _tool= DEFAULT_TOOL;

  if (_reset_tool)
    _reset_tool(this);

  _cursor= "";
  _handle_button.disconnect();
  _handle_motion.disconnect();
  _reset_tool.disconnect();

  if (notify)
    _owner->get_wb()->tool_changed(_view);
}


std::string ModelViewForm::get_tool_argument(const std::string &option)
{
  return _tool_args[option];
}


/** 
 ****************************************************************************
 * @brief Sets options for the selected tool
 *
 * Change the value of an option for the currently selected tool.
 *
 * @param option name of the option
 * @param value value of the option
 *****************************************************************************/
void ModelViewForm::set_tool_argument(const std::string &option, const std::string &value)
{
  _tool_args[option]= value;

  _tool_argument_changed.emit(option);
}


void ModelViewForm::set_zoom(double zoom)
{
  _model_view->zoom(zoom);
}


double ModelViewForm::get_zoom()
{
  return _model_view->zoom();
}


bec::Clipboard *ModelViewForm::get_clipboard()
{
  return _owner->get_grt_manager()->get_clipboard();
}



bool ModelViewForm::can_copy()
{
  return get_copiable_selection().count() > 0;
}


static void check_if_can_paste(WBComponent *compo, const grt::ObjectRef &object, bool *result)
{
  if (compo->can_paste_non_model_object(object))
    *result= true;
}


bool ModelViewForm::can_paste()
{
  std::list<grt::ObjectRef> data(get_clipboard()->get_data());
  WBContext *wb= _owner->get_wb();

  for (std::list<grt::ObjectRef>::iterator iter= data.begin();
      iter != data.end(); ++iter)
  {
    if (model_ObjectRef::can_wrap(*iter))
    {
      if (!wb->get_component_handling(model_ObjectRef::cast_from(*iter)))
        return false;
    }
    else
    {
      bool result= false;
      wb->foreach_component(sigc::bind<grt::ObjectRef,bool*>(sigc::ptr_fun(check_if_can_paste), *iter, &result));
      if (!result)
        return false;
    }
  }
  return !get_clipboard()->empty();
}


bool ModelViewForm::can_delete()
{
  return has_selection();
}


std::string ModelViewForm::get_edit_target_name()
{
  grt::ListRef<model_Object> sel(get_copiable_selection());

  if (sel.count()==0)
    return "";

  if (sel.count() == 1)
  {
    std::string name;

    name= sel[0]->name().c_str();
    if (name.empty() && sel[0]->has_member("caption"))
      name= sel[0]->get_string_member("caption");

    return fmt("'%s'", name.c_str());
  }
  else
    return fmt(_("%i Selected Figures"), sel.count());
}



void ModelViewForm::cut()
{
  grt::UndoManager *um= _owner->get_wb()->get_undo_manager();

  std::string edit_target_name= get_edit_target_name();

  um->begin_undo_group();
  copy();
  int count= delete_selection(true);
  um->end_undo_group();
  um->set_action_description(fmt(_("Cut %s"), edit_target_name.c_str()));

  _owner->get_wb()->show_status_text(fmt(_("%i object(s) cut."), count));
}


void ModelViewForm::copy()
{
  grt::ListRef<model_Object> selection(get_copiable_selection());
  bec::Clipboard *clip= get_clipboard();
  model_LayerRef target_layer(get_model_view()->rootLayer());
  int count= 0;
  grt::CopyContext copy_context(_owner->get_grt());

  clip->clear();
  for (size_t c= selection.count(), i= 0; i < c; i++)
  {
    WBComponent *compo= _owner->get_wb()->get_component_handling(selection.get(i));
    if (compo)
    {
      model_ObjectRef copy(compo->clone_object(selection.get(i), target_layer, copy_context, true));
      clip->append_data(copy);
      count++;
    }
  }
  clip->set_content_description(get_edit_target_name());

  copy_context.finish();
  clip->changed();

  _owner->get_wb()->show_status_text(fmt(_("%i object(s) copied."), count));
}


void ModelViewForm::clipboard_changed()
{
  _paste_offset= 0;
}


static void get_component_that_can_paste(WBComponent *compo, 
                                         const grt::ObjectRef &object, 
                                         WBComponent **result)
{
  if (compo->can_paste_non_model_object(object))
    *result= compo;
}


void ModelViewForm::paste()
{
  WBContext *wb= _owner->get_wb();
  int count= 0;

  _paste_offset+= 10;

  grt::CopyContext context(wb->get_grt());

  wb->get_grt()->begin_auto_undo();

  _model_view->selection().remove_all();

  std::list<grt::ObjectRef> data(get_clipboard()->get_data());
  for (std::list<grt::ObjectRef>::iterator iter= data.begin();
      iter != data.end(); ++iter)
  {
    model_LayerRef layer= _model_view->rootLayer();
    if (model_ObjectRef::can_wrap(*iter))
    {
      WBComponent *compo= wb->get_component_handling(model_ObjectRef::cast_from(*iter));

      if (!compo) continue;

      if (model_ObjectRef::can_wrap(*iter))
      {
        model_ObjectRef object(compo->clone_object(model_ObjectRef::cast_from(*iter), layer, context, false));
        if (object.is_valid())
        {
          if (object.is_instance(model_Figure::static_class_name()))
          {
            model_FigureRef figure(model_FigureRef::cast_from(object));
            figure->top(*figure->top() + _paste_offset);
            figure->left(*figure->left() + _paste_offset);
          }
          else if (object.is_instance(model_Layer::static_class_name()))
          {
            model_LayerRef layer(model_LayerRef::cast_from(object));
            layer->top(*layer->top() + _paste_offset);
            layer->left(*layer->left() + _paste_offset);
          }
          count++;

          _model_view->selection().insert(object);
        }
      }
    }
    else
    {
      WBComponent *compo= 0;
      wb->foreach_component(
        sigc::bind<grt::ObjectRef,WBComponent**>(sigc::ptr_fun(get_component_that_can_paste),
          *iter, &compo)); 
      if (compo)
      {
        compo->paste_non_model_object(this, *iter);
        count++;
      }
    }
  }
  context.finish();
  wb->get_grt()->end_auto_undo(fmt(_("Paste %s"), get_clipboard()->get_content_description().c_str()));

  _owner->get_wb()->show_status_text(fmt(_("%i object(s) pasted."), count));
}


void ModelViewForm::delete_selection()
{
  int count= delete_selection(false);
  _owner->get_wb()->show_status_text(fmt(_("%i object(s) deleted."), count));
}


int ModelViewForm::delete_selection(bool noquestions)
{
  grt::UndoManager *um= _owner->get_wb()->get_undo_manager();
  grt::ListRef<model_Object> selection= get_selection();

  std::vector<model_ObjectRef> objects;
  std::string edit_target_name= get_edit_target_name();

  um->begin_undo_group();

  for (size_t c= selection.count(), i= 0; i < c; i++)
  {
    if (selection.get(i).is_instance(model_Object::static_class_name()))
      objects.push_back(model_ObjectRef::cast_from(selection.get(i)));
  }

  std::map<std::string, bool> options;

  if (noquestions)
    options["deletedbobjects"]= true;
  else
  {
    std::string option;

    option= _owner->get_wb()->get_wb_options().get_string("workbench.physical:DeleteObjectConfirmation", "ask");
    if (option == "delete")
      options["deletedbobjects"]= true;
    else if (option == "keep")
      options["deletedbobjects"]= false;
  }
  for (size_t c= objects.size(), i= 0; i < c; i++)
    _owner->get_wb()->delete_object(objects[i], options);

  um->end_undo_group();
  um->set_action_description(fmt(_("Delete %s"), edit_target_name.c_str()));

  return objects.size();
}


grt::ListRef<model_Object> ModelViewForm::get_selection()
{
  return _model_view->selection();
}


grt::ListRef<model_Object> ModelViewForm::get_copiable_selection()
{
  grt::ListRef<model_Object> sel(_model_view->selection());
  grt::ListRef<model_Object> copiable(_owner->get_grt());

  for (size_t c= sel.count(), i= 0; i < c; i++)
  {
    if (!sel.get(i).is_instance(model_Connection::static_class_name()))
      copiable.insert(sel[i]);
  }
  return copiable;
}


bool ModelViewForm::has_selection()
{
  return _model_view->selection().count() > 0;
}

void ModelViewForm::focus_and_make_visible(const model_ObjectRef &object, bool select)
{
  mdc::CanvasItem *item= 0;

  if (object.is_instance(model_Figure::static_class_name()))
  {
    item= model_FigureRef::cast_from(object)->get_data()->get_canvas_item();
  }
  else if (object.is_instance(model_Connection::static_class_name()))
  {
    item= model_ConnectionRef::cast_from(object)->get_data()->get_canvas_item();
  }
  else if (object.is_instance(model_Layer::static_class_name()))
  {
    item= model_LayerRef::cast_from(object)->get_data()->get_area_group();
  }
 

  if (item)
  {
    mdc::CanvasView *view= item->get_view();
    mdc::Rect bounds= item->get_root_bounds();
    mdc::Rect viewport= view->get_viewport();

    if (!mdc::bounds_contain_bounds(viewport, bounds))
    {
      mdc::Point offset= viewport.pos;

      // get the offset that will move the viewport the least to bring the item inside it

      if (bounds.xmin() < viewport.xmin())
        offset.x= bounds.xmin() - 20;
      else if (bounds.xmax() > viewport.xmax())
        offset.x= bounds.xmax() - viewport.width();

      if (bounds.ymin() < viewport.ymin())
        offset.y= bounds.ymin() - 20;
      else if (bounds.ymax() > viewport.ymax())
        offset.y= bounds.ymax() - viewport.height();

      view->set_offset(offset);
    }
    view->focus_item(item);
    if (select)
      view->get_selection()->set(item);
  }
}


void ModelViewForm::set_cursor(const std::string &cursor)
{
  _cursor= cursor;
}



model_LayerRef ModelViewForm::get_layer_at(const mdc::Point &pos, mdc::Point &offset)
{
  model_LayerRef layer;
  mdc::AreaGroup *ag= 0;

  mdc::CanvasItem *item= _view->get_item_at(pos);

  if (!item)
  {
    offset= pos;
    return _model_view->rootLayer();
  }
  item= item->get_toplevel();


  while (item && (ag= dynamic_cast<mdc::AreaGroup*>(item))==0)
    item= item->get_parent();

  if (ag)
  {
    for (size_t c= _model_view->layers().count(), i= 0; i < c; i++)
    {
      layer= _model_view->layers().get(i);

      if (ag == layer->get_data()->get_area_group())
      {
        offset= ag->convert_point_from(pos, 0);

        return layer;
      }
    }
  }

  return _model_view->rootLayer();
}


model_LayerRef ModelViewForm::get_layer_bounding(const mdc::Rect &rect, mdc::Point &offset)
{
  for (ssize_t i= _model_view->layers().count()-1; i >= 0; --i)
  {
    model_LayerRef layer(_model_view->layers()[i]);

    if (layer->get_data()->get_area_group() &&
      mdc::bounds_contain_bounds(layer->get_data()->get_area_group()->get_root_bounds(), rect))
    {
      return layer;
    }
  }
  return model_LayerRef();
}


model_ObjectRef ModelViewForm::get_object_at(const mdc::Point &pos)
{
  mdc::CanvasItem *item= _view->get_item_at(pos);

  if (!item)
    return model_ObjectRef();

  std::string id= item->get_tag();
  if (id.empty())
    return model_ObjectRef();

  model_ObjectRef object;

  object= grt::find_object_in_list(_model_view->figures(), id);
  if (object.is_valid())
    return object;

  object= grt::find_object_in_list(_model_view->layers(), id);
  if (object.is_valid())
    return object;

  object= grt::find_object_in_list(_model_view->connections(), id);
  if (object.is_valid())
    return object;

  return model_ObjectRef();
}

grt::UndoManager *ModelViewForm::get_undo_manager()
{
  return _owner->get_wb()->get_undo_manager(); 
}


bool ModelViewForm::relocate_selection()
{
  bool relocated= false;
  grt::UndoManager *um= get_undo_manager();
  grt::ListRef<model_Figure> figures(_model_view->figures());

  for (size_t c= figures.count(), i= 0; i < c; i++)
  {
    model_FigureRef figure(figures[i]);
    mdc::CanvasItem *item= figure->get_data()->get_canvas_item();

    if (!item) continue;

    mdc::Rect bounds(item->get_root_bounds());
    mdc::Point offset;
    
    model_LayerRef layer(get_layer_bounding(bounds, offset));

    if (layer.is_valid())
    {
      // figure is in some layer
      if (layer != figure->layer())
      {
        if (figure->layer().is_valid())
        {
          um->add_undo(new grt::UndoListRemoveAction(figure->layer()->figures(), figure));
          figure->layer()->figures().remove_value(figure);
        }
        um->add_undo(new grt::UndoObjectChangeAction(figure, "layer"));
        figure->layer(layer);
        um->add_undo(new grt::UndoListInsertAction(layer->figures()));
        layer->figures().insert(figure);
        relocated= true;
      }
      else
      {
        mdc::AreaGroup *parent= layer->get_data()->get_area_group();
        // if figure stays in same layer as before, make sure that it has correct canvas parent
        if (parent && item->get_parent() != parent)
        {
          mdc::Point pos= item->get_root_position() - parent->get_root_position();

          parent->add(item);
          item->move_to(pos);
        }
      }
    }
    else
    {
      // figure is outside of any layer

      if (figure->layer() != _model_view->rootLayer())
      {
        if (figure->layer().is_valid())
        {
          um->add_undo(new grt::UndoListRemoveAction(figure->layer()->figures(), figure));
          figure->layer()->figures().remove_value(figure);
        }
        um->add_undo(new grt::UndoObjectChangeAction(figure, "layer"));
        figure->layer(_model_view->rootLayer());
        um->add_undo(new grt::UndoListInsertAction(_model_view->rootLayer()->figures()));
        _model_view->rootLayer()->figures().insert(figure);
        relocated= true;
      }
    }
  }
  return relocated;
}



bool ModelViewForm::accepts_drop(int x, int y, const std::string &type, const std::list<GrtObjectRef> &objects)
{
  return _owner->accepts_drop(this, x, y, type, objects);
}


bool ModelViewForm::accepts_drop(int x, int y, const std::string &type, const std::string &text)
{
  return _owner->accepts_drop(this, x, y, type, text);
}


bool ModelViewForm::perform_drop(int x, int y, const std::string &type, const std::list<GrtObjectRef> &objects)
{
  return _owner->perform_drop(this, x, y, type, objects);
}

bool ModelViewForm::perform_drop(int x, int y, const std::string &type, const std::string &text)
{
  return _owner->perform_drop(this, x, y, type, text);
}




mdc::Layer *ModelViewForm::get_floater_layer()
{
  return _floater_layer;
}


void ModelViewForm::add_floater(Floater *floater)
{
  mdc::Point pos;

  pos.x= _view->get_viewport().xmax() - 200;
  pos.y= _view->get_viewport().ymin() + 20;

  floater->move_to(pos);
  _floater_layer->add_item(floater);
}


/* unused
void ModelViewForm::hover_in_figure(mdc::CanvasItem *item, mdc::Point pos)
{
}



void ModelViewForm::hover_out_figure(mdc::CanvasItem *item)
{
}
*/
