
#include <mforms/simpleform.h>
#include <mforms/mforms.h>

#include <stdlib.h>
#include <string.h>

#include <sstream>
#include <stdexcept>

#define MIN_CONTROL_WIDTH 200
#define FORM_MARGIN  12


using namespace mforms;

SimpleForm::SimpleForm(const std::string &title, const std::string &ok_caption)
{
  ControlFactory *f= ControlFactory::get_instance();

  _caption_width= 0;
  _control_width= 0;
  _title_width= 0;

  _ok_button= 0;
  _cancel_button= 0;

  _ok_caption= ok_caption;

  _form= new Form();
  _form->set_title(title);
  
  _content= new Panel(false);
  _form->set_content(_content);
}


SimpleForm::~SimpleForm()
{
  for (std::list<Row>::iterator iter= _rows.begin(); iter != _rows.end(); ++iter)
  {
    delete iter->caption;
    delete iter->control;
  }
  delete _ok_button;
  delete _cancel_button;
  delete _form;
}


static std::string parse_newlines(const std::string &str)
{
  std::string result;
  const char *begin= str.data();
  const char *pos;

  while ((pos= strstr(begin, "\\n")) != NULL)
  {
    result.append(begin, pos).append("\n");
    begin= pos+2;
  }
  result.append(begin);

  return result;
}


/*
 * Example:
 *   label;Stuff Here\nOptionally in multiple\nlines.
 *   textentry;name;caption;default
 *   checkbox;name;caption;1/0
 *   textarea;name;caption;rows;default
 *   select;name;caption;item1,item2,item3;default_index
 */
void SimpleForm::parse_definition(const std::string &definition)
{
  std::stringstream strm(definition, std::stringstream::in);
  char line[4*1024];
  char arg[4*1024];

  while (!strm.eof())
  {
    char type[20];

    strm.getline(line, sizeof(line));

    if (!*line) continue;

    std::stringstream toks(line, std::stringstream::in);

    toks.getline(type, sizeof(type), ';');
    if (strcmp(type, "label")==0)
    {
      toks.getline(arg, sizeof(arg));
      add_label(parse_newlines(arg), false);
    }
    else if (strcmp(type, "textentry")==0)
    {
      std::string name, caption, defaultv;

      toks.getline(arg, sizeof(arg), ';');
      name= arg;
      toks.getline(arg, sizeof(arg), ';');
      caption= arg;
      toks.getline(arg, sizeof(arg), ';');
      defaultv= arg;
      
      add_text_entry(name, caption, defaultv);
    }
    else if (strcmp(type, "checkbox")==0)
    {
      std::string name, caption, defaultv;

      toks.getline(arg, sizeof(arg), ';');
      name= arg;
      toks.getline(arg, sizeof(arg), ';');
      caption= arg;
      toks.getline(arg, sizeof(arg), ';');
      defaultv= arg;
      
      add_checkbox(name, caption, defaultv=="1");
    }
    else if (strcmp(type, "textarea")==0)
    {
      std::string name, rows, caption, defaultv;

      toks.getline(arg, sizeof(arg), ';');
      name= arg;
      toks.getline(arg, sizeof(arg), ';');
      caption= arg;
      toks.getline(arg, sizeof(arg), ';');
      rows= arg;
      toks.getline(arg, sizeof(arg), ';');
      defaultv= parse_newlines(arg);
      
      add_text_area(name, caption, atoi(rows.c_str()), defaultv);
    }
    else if (strcmp(type, "select")==0)
    {
      std::string name, caption, items, defaulti;
      std::list<std::string> items_list;

      toks.getline(arg, sizeof(arg), ';');
      name= arg;
      toks.getline(arg, sizeof(arg), ';');
      caption= arg;
      toks.getline(arg, sizeof(arg), ';');
      items= arg;
      toks.getline(arg, sizeof(arg), ';');
      defaulti= parse_newlines(arg);

      std::stringstream is(items, std::stringstream::in);
      char *s= new char[items.size()];
      while (!is.eof())
      {
        is.getline(s,items.size(), ',');
        items_list.push_back(s);
      }
      delete []s;
      
      add_select(name, caption, items_list, defaulti.empty()?-1:atoi(defaulti.c_str()));
    }
    else
      throw std::runtime_error(std::string("invalid simpleform control type: ").append(type));
  }
}


void SimpleForm::add_label(const std::string &text, bool bold)
{
  Label *l= new Label();
  l->set_text(text);
  _content->add(l);

  _title_width= std::max(_title_width, l->get_preferred_width());

  Row row;
  row.caption= l;
  row.control= 0;
  row.spacing= 12;
  row.fullwidth= false;
  _rows.push_back(row);
}


void SimpleForm::add_file_picker(const std::string &name, const std::string &caption, 
  const std::string &default_value)
{
}


void SimpleForm::add_text_entry(const std::string &name, const std::string &caption, 
  const std::string &default_value)
{
  Label *l= 0;

  if (!caption.empty())
  {
    l= new Label();
    l->set_text(caption);
    l->set_text_align(BottomRight);
    _content->add(l);

    int w= l->get_preferred_width();
    _caption_width= std::max(w, _caption_width);
  }

  TextEntry *t= new TextEntry();
  t->set_value(default_value);
  t->set_name(name);
  _content->add(t);

  _control_width= std::max(_control_width, t->get_preferred_width());

  Row row;
  row.caption= l;
  row.control= t;
  row.spacing= 4;
  row.fullwidth= false;
  _rows.push_back(row);
}


void SimpleForm::add_text_area(const std::string &name, const std::string &caption, int rows,
  const std::string &default_value)
{
  Label *l= 0;

  if (!caption.empty())
  {
    l= new Label();
    l->set_text(caption);
    l->set_text_align(BottomRight);
    _content->add(l);

    int w= l->get_preferred_width();
    _caption_width= std::max(w, _caption_width);
  }

  TextBox *t= new TextBox();
  t->set_value(default_value);
  t->set_name(name);
  t->set_size(t->get_preferred_width(), t->get_preferred_height()*rows);
  _content->add(t);

  _control_width= std::max(_control_width, t->get_preferred_width());

  Row row;

  row.caption= l;
  row.control= 0;
  row.spacing= 2;
  row.fullwidth= false;
  _rows.push_back(row);

  row.caption= 0;
  row.control= t;
  row.spacing= 12;
  row.fullwidth= true;
  _rows.push_back(row);
}


void SimpleForm::add_checkbox(const std::string &name, const std::string &caption, 
  bool default_value)
{
  CheckBox *t= new CheckBox();
  t->set_text(caption);
  t->set_active(default_value);
  t->set_name(name);
  _content->add(t);

  _control_width= std::max(_control_width, t->get_preferred_width());

  Row row;
  row.caption= 0;
  row.control= t;
  row.spacing= 4;
  row.fullwidth= false;
  _rows.push_back(row);
}


void SimpleForm::add_select(const std::string &name, const std::string &caption,
      const std::list<std::string> &items, int default_index)
{
  Label *l= 0;

  if (!caption.empty())
  {
    l= new Label();
    l->set_text(caption);
    l->set_text_align(BottomRight);
    _content->add(l);

    int w= l->get_preferred_width();
    _caption_width= std::max(w, _caption_width);
  }

  Selector *t= new Selector();
  t->set_selected(default_index);
  t->add_items(items);
  t->set_name(name);
  _content->add(t);

  _control_width= std::max(_control_width, t->get_preferred_width());

  Row row;
  row.caption= l;
  row.control= t;
  row.spacing= 4;
  row.fullwidth= false;
  _rows.push_back(row);
}


void SimpleForm::layout()
{
  int x, y;
  int form_width, form_height;
  int control_width= _control_width;

  x= FORM_MARGIN;
  y= FORM_MARGIN;

  if (_title_width > _control_width + _caption_width)
    control_width= _title_width - _caption_width - 4;

  if (control_width < MIN_CONTROL_WIDTH)
    control_width= MIN_CONTROL_WIDTH;

  for (std::list<Row>::const_iterator iter= _rows.begin(); iter != _rows.end(); ++iter)
  {
    int h= 0;

    if (iter->caption)
    {
      iter->caption->set_position(x, y+2);
      h= iter->caption->get_preferred_height();

      if (iter->control)
        iter->caption->set_size(_caption_width, h);
      else
        iter->caption->set_size(iter->caption->get_preferred_width(), h);
    }

    if (iter->control)
    {
      if (iter->fullwidth)
      {
        iter->control->set_position(x, y);
        iter->control->set_size(_caption_width + control_width + 4, iter->control->get_height());
      }
      else
      {
        iter->control->set_position(x + _caption_width + 4, y);
        iter->control->set_size(control_width, iter->control->get_preferred_height());
      }
      h= std::max(iter->control->get_height(), h);
    }

    y+= h + iter->spacing;
  }

  form_height= y;
  form_width= FORM_MARGIN*2 + _caption_width + 4 + control_width;

  _ok_button= new Button();
  _ok_button->set_text(_ok_caption);
  _ok_button->set_size(_ok_button->get_preferred_width(), _ok_button->get_preferred_height());
  _content->add(_ok_button);

  _cancel_button= new Button();
  _cancel_button->set_text("Cancel");
  _cancel_button->set_size(_cancel_button->get_preferred_width(), _cancel_button->get_preferred_height());
  _content->add(_cancel_button);

  _ok_button->set_position(form_width - FORM_MARGIN - _ok_button->get_width() - 8 - _cancel_button->get_width(),
    form_height + FORM_MARGIN);
  _cancel_button->set_position(form_width - FORM_MARGIN - _cancel_button->get_width(),
    form_height + FORM_MARGIN);

  form_height += _ok_button->get_height() + FORM_MARGIN*2;

  _form->set_size(form_width, form_height);
}


bool SimpleForm::show()
{
  layout();

  return _form->run_modal(_ok_button, _cancel_button);
}


std::map<std::string,Control*> SimpleForm::get_controls()
{
  std::map<std::string,Control*> ctls;

  for (std::list<Row>::const_iterator iter= _rows.begin(); iter != _rows.end(); ++iter)
  {
    if (iter->control)
      ctls[iter->control->get_name()]= iter->control;
  }

  return ctls;
}


std::string SimpleForm::get_string_value(const std::string &name)
{
  Control *control= dynamic_cast<Control*>(_content->find_subview(name));
  if (control)
    return control->get_string_value();

  return "";
}


bool SimpleForm::get_bool_value(const std::string &name)
{
  Control *control= dynamic_cast<Control*>(_content->find_subview(name));
  if (control)
    return control->get_bool_value();

  return false;
}


int SimpleForm::get_int_value(const std::string &name)
{
  Control *control= dynamic_cast<Control*>(_content->find_subview(name));
  if (control)
    return control->get_int_value();

  return 0;
}

